/*
 * Created on 16.09.2004
 *
 * (C) Peter Weigerber <peter.weissgerber@ku-eichstaett.de>
 * 
 * For license information see the LICENSE file that is shipped
 * with the source code.
 * 
 * 
 */
package org.eposoft.epos.eposee.pixmap.data;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Vector;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

//import org.apache.log4j.Logger;
import org.eposoft.epos.util.comparator.FilenameComparator;



/** This class holds a pixelmap.
 * The implementation aims at being low-memory rather than on high speed.
 * A pixelmap is a graph with binary couplings between items.
 * Items are nodes and relations between them are directed edges. 
 * Each edge has two values annotated: the support and the confidence 
 * of the coupling.
 * 
 * @author Peter Weissgerber <peter.weissgerber@ku-eichstaett.de>
 * 
 */
public class SparePixMap {

	private String mapstring[] = null;

	private int size;

	private Hashtable couplings;

	//private static Logger log = Logger.getLogger(SparePixMap.class);

	private static int lastChar = -1;

	/** Creates an empty pixelmap.
	 * 
	 * @param _size size of the pixelmap
	 * 
	 */
	public SparePixMap(int _size) {
		size = _size;
		mapstring = new String[size];
		couplings = new Hashtable();
	}

	/** Set mapping "item IDs --> item names".
	 * 
	 * @param _mappings mappings as string array
	 */
	public void setMappings(String[] _mappings) {
		if (_mappings.length == size) {
			mapstring = _mappings;
		} else {
			//log.error("Mappings do not fit size  - mappings not set.");
		}
	}

	/**
	 * Retrieve size
	 * 
	 * @return size of the pixmap
	 */
	public int size() {
		return size;
	}

	/** Add a coupling to the system. 
	 * If a coupling of the spezified items is
	 * already in the system, it will be overwritten with the new values
	 * 
	 * @param _ante item ID of antecedent
	 * @param _cons item ID of consequent
	 * @param _sup  support count of the coupling
	 * @param _conf confidence of the coupling
	 */
	public void putCoupling(int _ante, int _cons, int _sup, double _conf) {
		if ((_ante < size) && (_cons < size)) {
			BinaryCouplingI cp = new BinaryCouplingI(_ante, _cons, _sup, _conf);
			PixMapCoordinate co = new PixMapCoordinate(_ante, _cons);
			couplings.put(co, cp);
		} else {
			//log.error("Position does not exist.");
		}
	}

	/** Add a coupling to the system.
	 * 
	 * @param _coupling binary coupling to be added
	 */
	public void putCoupling(BinaryCouplingI _coupling) {
		if ((_coupling.getAntecedent() < size)
				&& (_coupling.getConsequent() < size)) {
			PixMapCoordinate co = new PixMapCoordinate(_coupling
					.getAntecedent(), _coupling.getConsequent());
			couplings.put(co, _coupling);
		} else {
			//log.error("Position does not exist.");
		}
	}

	/** Get an array containing all couplings in this pixelmap.
	 * 
	 * @return array of couplings
	 */
	public BinaryCouplingI[] getCouplings() {
		Enumeration enum = couplings.elements();
		Vector cv = new Vector();
		while (enum.hasMoreElements()) {
			cv.add(enum.nextElement());
		}
		BinaryCouplingI[] coups = new BinaryCouplingI[cv.size()];
		for (int i = 0; i < cv.size(); i++) {
			coups[i] = (BinaryCouplingI) cv.get(i);
		}
		return coups;
	}
	
	/** Get the coupling at a certain coordinate of the pixel map
	 * 
	 * @param x x position
	 * @param y y position
	 * @return the BinaryCoupling at (x,y) or null, if there is no coupling
	 */
	public BinaryCouplingI getCouplingAt(int x, int y) {
		PixMapCoordinate coord = new PixMapCoordinate(x, y);
		if (couplings.containsKey(coord)) {
			return (BinaryCouplingI)couplings.get(coord);
		}
		else {
			return null;
		}
	}

	/** Get mappings as string array. 
	 * Note: The array is cached and not newly computed every time
	 * 
	 * @return string array with mappings
	 */
	public String[] getMappings() {
		return mapstring;
	}

	/**
	 * Sort pixelmap according to hierarchical order
	 *  
	 */
	public void sort() {
		String[] items = (String[]) mapstring.clone();
		Arrays.sort(items, new FilenameComparator());

		int[] sortarray = new int[items.length];
		for (int i = 0; i < items.length; i++) {
			boolean found = false;
			for (int j = 0; (j < items.length) && (!found); j++) {
				if (items[j].equals(mapstring[i])) {
					sortarray[i] = j;
					found = true;
				}
			}
		}

		Hashtable ht = new Hashtable();

		BinaryCouplingI[] coups = this.getCouplings();
		for (int i = 0; i < coups.length; i++) {
			int ante = sortarray[coups[i].getAntecedent()];
			int cons = sortarray[coups[i].getConsequent()];
			coups[i].setAntecedent(ante);
			coups[i].setConsequent(cons);
			ht.put(new PixMapCoordinate(ante, cons), coups[i]);
		}
		couplings = ht;
		mapstring = items;
	}

	/**
	 * Implements a kind of "-" operator: All couplings in the argument pixelmap
	 * will be removed from the original pixelmap.
	 * 
	 * @param pix
	 *                   argument pixelmap
	 */
	public void getDifferencesTo(SparePixMap pix) {
		BinaryCouplingI[] coups1 = this.getCouplings();
		BinaryCouplingI[] coups2 = pix.getCouplings();
		String[] labels2 = pix.getMappings();

		Hashtable ht = new Hashtable();

		String antestr1 = "";
		String consstr1 = "";
		String antestr2 = "";
		String consstr2 = "";
		int ante = -1;
		int cons = -1;
		// TODO this implementation is in O(n^2). can we do it better?
		for (int i = 0; i < coups1.length; i++) {
			boolean found = false;
			ante = coups1[i].getAntecedent();
			cons = coups1[i].getConsequent();
			antestr1 = mapstring[ante];
			consstr1 = mapstring[cons];
			for (int j = 0; ((j < coups2.length) && (!found)); j++) {
				antestr2 = labels2[coups2[j].getAntecedent()];
				consstr2 = labels2[coups2[j].getConsequent()];
				if ((antestr1.equals(antestr2)) && (consstr1.equals(consstr2))) {
					// coupling is in pix ==> does not pass filter
					found = true;
				}
			}
			if (!found) {
				ht.put(new PixMapCoordinate(ante, cons), coups1[i]);
			}
		}

		couplings = ht;

	}
	
	public void exportToGML(String filename) throws IOException {
		BufferedWriter wr;
		
		if (filename.toLowerCase().endsWith(".gml")) {
			wr = new BufferedWriter(new FileWriter(filename));
		}
		else if (filename.toLowerCase().endsWith(".gml.gz")) {
			wr = new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(new FileOutputStream(filename))));
		}
		else {
			filename += ".gml.gz";
			wr = new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(new FileOutputStream(filename))));
		}
		
		//wr = new BufferedWriter(new FileWriter(filename));
		wr.write("graph [\n");
		
		wr.write("    comment \"GML output\"\n");
		wr.write("    directed 1\n");
		wr.write("    id 1\n");
		wr.write("    label \"Testlabel FIXME\"\n");
		
		for (int i=0;i<mapstring.length;i++) {
			wr.write("    node [\n");
			wr.write("        id "+i+"\n");
			wr.write("        label \""+mapstring[i]+"\"\n");
			wr.write("    ]\n");			
		}
		
		BinaryCouplingI[] coups = this.getCouplings();
		
		for (int i=0;i<coups.length;i++) {
			wr.write("    edge [\n");
			wr.write("        label \"");
			wr.write(mapstring[coups[i].getAntecedent()]);
			wr.write(" -> ");
			wr.write(mapstring[coups[i].getConsequent()]);
			wr.write("\"\n");
			wr.write("        source "+coups[i].getAntecedent()+"\n");
			wr.write("        target "+ coups[i].getConsequent() + "\n");
			wr.write("        support "+coups[i].getSupportCount()+"\n");
			wr.write("        confidence "+coups[i].getConfidence()+"\n");
			wr.write("    ]\n");
		}
				
		wr.write("]\n");
		wr.close();
	}

    /**
     * Import pixmap from TXT files
     * 
     * @param supfile
     *                   file containing support values
     * @param conffile
     *                   file containing confidence values
     * @param mapfile
     *                   file containing item names and mappings to positions in matrix
     * @param filterfile
     *                   file containing item filters, use "" for no filter file
     * @return PixMap object created from these files
     * @throws IOException
     */
    public static SparePixMap fromTXTFiles(String supfile, String conffile,
            String mapfile, String filterfile) throws IOException {

        BufferedReader confr = null;
        if (conffile.toLowerCase().endsWith(".gz")) {
            confr = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(conffile))));
        } else {
            confr = new BufferedReader(new FileReader(conffile));
        }
        
        SparePixMap mypix = new SparePixMap(1000);
        String[] tmp_mappings = new String[1000];
        for (int i = 0; i < tmp_mappings.length; i++) {
            tmp_mappings[i] = "" + (i+1);
        }
        mypix.setMappings(tmp_mappings);

        String confline = "";

        int linenr = -1;

        while ((confline = confr.readLine()) != null) {
            linenr++;
            String[] cvals = confline.split(" ");
            if (cvals.length==3) {
                int itemx = Integer.parseInt(cvals[0]);
                int itemy = Integer.parseInt(cvals[1]);

                double actualval = Double.parseDouble(cvals[2]);
                mypix.putCoupling(itemx, itemy, 1, actualval);
            }
        }
        confr.close();
        return mypix;
    }

    
    /**
	 * Import pixmap from CVS files
	 * 
	 * @param supfile
	 *                   file containing support values
	 * @param conffile
	 *                   file containing confidence values
	 * @param mapfile
	 *                   file containing item names and mappings to positions in matrix
	 * @param filterfile
	 *                   file containing item filters, use "" for no filter file
	 * @return PixMap object created from these files
	 * @throws IOException
	 */
	public static SparePixMap fromCSVFiles(String supfile, String conffile,
			String mapfile, String filterfile) throws IOException {

		HashSet passing = new HashSet();

		// Loading filters if specified
		boolean filters = true;
		if (filterfile.equals("")) {
			filters = false;
			//log.debug("Filter disabled (no file specified)");
		} else {
			try {
				BufferedReader filterr = new BufferedReader(new FileReader(
						filterfile));
				//log.debug("Reading filter file " + filterfile);
				String filterline = "";
				while ((filterline = filterr.readLine()) != null) {
					String[] flparsed = filterline.split("-");
					int begln = Integer.parseInt(flparsed[0]);
					int endln = begln;
					if (flparsed.length>1) {
						endln = Integer.parseInt(flparsed[1]);
					}
					for (int i = begln; i <= endln; i++) {
						passing.add(new Integer(i));
					}
				}
				filterr.close();
				System.out.println("Filter enabled");
			} catch (FileNotFoundException e2) {
				filters = false;
				System.out.println("Filter disabled (filter file not found)");
			}
		}

		//log.debug("Reading mapping file " + mapfile);
		Hashtable filtermappings = new Hashtable();
		BufferedReader mapr = null;
		if (mapfile.toLowerCase().endsWith(".gz")) {
			mapr = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(mapfile))));
		} else {
			mapr = new BufferedReader(new FileReader(mapfile));
		}
		Vector _mappings = new Vector();
		int itemidold = 0;
		String mapline = "";
		int itemidnew = -1;
		while ((mapline = mapr.readLine()) != null) {
			itemidold++;
			if ((!filters)
					|| ((filters) && (passing.contains(new Integer(itemidold))))) {
				itemidnew++;
				int split = mapline.indexOf(' ');
				_mappings.add(mapline.substring(split + 1));
				filtermappings.put(new Integer(itemidold), new Integer(
						itemidnew));
			}
		}
		mapr.close();

		SparePixMap mypix = new SparePixMap(itemidnew + 1);
		String[] tmp_mappings = new String[_mappings.size()];
		for (int i = 0; i < tmp_mappings.length; i++) {
			tmp_mappings[i] = (String) _mappings.get(i);
		}
		mypix.setMappings(tmp_mappings);

		//log.debug("Reading support file " + supfile + " and confidence file "
		//		+ conffile);

		BufferedReader confr = null;
		if (conffile.toLowerCase().endsWith(".gz")) {
			confr = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(conffile))));
		} else {
			confr = new BufferedReader(new FileReader(conffile));
		}
		
		BufferedReader supr = null;
		if (supfile.toLowerCase().endsWith(".gz")) {
			supr = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(supfile))));
		} else {
			supr = new BufferedReader(new FileReader(supfile));
		}

		// normally confidence- and supportfile should have exactly the same
		// number of
		// columns and lines - if not the files are corrupted and this alg. will
		// fail

		String confline = "";
		String supline = "";

		int linenr = -1;
		int percent;
		int lastp = 0;

		//PixMapCoordinate coord = new PixMapCoordinate(0, 0);

		while ((confline = confr.readLine()) != null) {
			supline = supr.readLine();
			linenr++;
			percent = (int) (((double) linenr / (double) itemidold) * 100);
			if (percent % 5 == 0) {
				if (percent > lastp) {
					System.out.println(percent + "% done");
				}
				lastp = percent;
			}
			if ((!filters)
					|| ((filters) && (passing.contains(new Integer(linenr + 1))))) {
				String[] cvals = confline.split(";");
				String[] svals = supline.split(";");
				for (int i = 0; i < cvals.length; i++) {
					if ((!filters)
							|| ((filters) && (passing.contains(new Integer(
									i + 1))))) {

						int support = Integer.parseInt(svals[i]);

						if (support > 0) {
							int itemx = ((Integer) filtermappings
									.get(new Integer(linenr + 1))).intValue();
							int itemy = ((Integer) filtermappings
									.get(new Integer(i + 1))).intValue();

							double actualval = Double.parseDouble(cvals[i]);
							//short value = (short) (actualval * 1000);

							mypix.putCoupling(itemx, itemy, support, actualval);

						}
					}
				}
			}
		}
		supr.close();
		confr.close();

		return mypix;
	}

	/**
	 * Import pixmap from GML file
	 * 
	 * @param filename
	 *                   filename
	 * @return PixMap object created from this file
	 * @throws FileNotFoundException
	 *                   thr
	 */
	public static SparePixMap fromGMLFile(String filename, String filterfile) throws IOException {

		BufferedReader fr=null;
		
		if (filename.toLowerCase().endsWith(".gz")) {
			fr = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(filename))));
		}
		else {
			fr = new BufferedReader(new FileReader(filename));
		}

		
		Vector passing = new Vector();
		boolean filter = false;
		if (!filterfile.equals("")) {
			System.out.println("filter file = "+filterfile);
			try {
				BufferedReader filterr = new BufferedReader(new FileReader(
						filterfile));
				//log.debug("Reading filter file " + filterfile);
				String filterline = "";
				while ((filterline = filterr.readLine()) != null) {
					String[] flparsed = filterline.split("-");
					int begln = Integer.parseInt(flparsed[0]);
					int endln = begln;
					if (flparsed.length>1) {
						endln = Integer.parseInt(flparsed[1]);
					}
					for (int i = begln; i <= endln; i++) {
						passing.add(new Integer(i));
					}
				}
				filterr.close();
				filter = true;
				//log.debug("Filter enabled");
			}
			catch (IOException e) {
				filter = false;
				//log.warn("Filter file not found - disabling filter");
			}
		}
		
		
		// temp. variables
		Hashtable itemtable = new Hashtable();
		Vector mappings = new Vector();
		Vector couplings = new Vector();

		boolean end = false;

		while (!end) {
			String mytoken = getNextToken(fr);
			//log.debug("token <"+mytoken+">");
			if (mytoken.equals("")) {
				System.out.println("<EOF>");
				end = true;
			} else {
				if (mytoken.equals("graph")) {
					//log.debug("graph begin");
					// next token should be [
					mytoken = getNextToken(fr);
					while (!mytoken.equals("]")) {
						mytoken = getNextToken(fr);
						if (mytoken.equals("node")) {
							readNode(fr, itemtable, mappings, filter, passing);
						} else if (mytoken.equals("\"")) {
							readString(fr);
						}

						else if (mytoken.equals("edge")) {

							BinaryCouplingI coupl = readEdge(fr, itemtable);
							if (coupl!=null) {
								couplings.add(coupl);
							}

						}

						//mytoken=getNextToken(fr);

					}
				}

				//System.out.println("Token: <"+mytoken+">");
			}
		}

		//log.debug("Reading finished");
		SparePixMap mypix = new SparePixMap(mappings.size());
		String[] mapstr = new String[mappings.size()];
		for (int i = 0; i < mappings.size(); i++) {
			mapstr[i] = (String) mappings.get(i);
		}
		mypix.setMappings(mapstr);
		for (int i = 0; i < couplings.size(); i++) {
			mypix.putCoupling((BinaryCouplingI) couplings.get(i));
		}
		mypix.sort();
		return mypix;
	}

	////////////////////// STATIC HELPER METHODS
	////////////////////// I KNOW THIS IS A HACK, BUT AT THE MOMENT IT WORKS...

	/**
	 * @param fr
	 * @param itemtable
	 * @return
	 */
	private static BinaryCouplingI readEdge(BufferedReader fr,
			Hashtable itemtable) {
		String token = "";
		String tmp = "";

		int tmp_edgefrom = -1;
		int tmp_edgeto = -1;
		int tmp_sup = 1;
		double tmp_conf = 1.0;

		token = getNextToken(fr);
		if (token.equals("[")) {
			boolean end = false;
			while (!end) {
				token = getNextToken(fr);
				if (token.equals("source")) {
					tmp = getNextToken(fr);
					tmp_edgefrom = Integer.parseInt(tmp);
				} else if (token.equals("target")) {
					tmp = getNextToken(fr);
					tmp_edgeto = Integer.parseInt(tmp);
				} else if (token.equals("support")) {
					tmp = getNextToken(fr);
					tmp_sup = Integer.parseInt(tmp);
				} else if (token.equals("confidence")) {
					tmp = getNextToken(fr);
					tmp_conf = Double.parseDouble(tmp);
				} else if (token.equals("\"")) {
					readString(fr);
				} else if (token.equals("[")) {
					ignoreSection(fr);
				} else if (token.equals("]")) {
					end = true;

					//log.debug("[edge from=" + tmp_edgefrom + " to="
					//		+ tmp_edgeto + "]");
					int xpos=-1;
					int ypos=-1;
					try {
						xpos = ((Integer) itemtable.get(new Integer(
							tmp_edgefrom))).intValue();
						ypos = ((Integer) itemtable
							.get(new Integer(tmp_edgeto))).intValue();
					}
					catch (NullPointerException e) {
						System.out.println("[edge filtered out]");
						return null;
					}
					return (new BinaryCouplingI(xpos, ypos, tmp_sup, tmp_conf));

				}
			}
		}
		return null;
	}

	/**
	 * @param fr
	 * @param itemtable
	 * @return
	 */
	private static void readNode(BufferedReader fr, Hashtable itemtable,
			Vector mappings, boolean filter, Vector passing) {
		String token = "";
		String tmp = "";
		int tmp_nodeid = -1;
		String tmp_nodelabel = "";

		token = getNextToken(fr);
		if (token.equals("[")) {
			boolean end = false;
			while (!end) {
				token = getNextToken(fr);
				if (token.equals("id")) {
					// read id
					tmp = getNextToken(fr);
					tmp_nodeid = Integer.parseInt(tmp);
				} else if (token.equals("label")) {
					token = getNextToken(fr);
					if (token.equals("\"")) {
						tmp_nodelabel = readString(fr);
					} else {
						tmp_nodelabel = token;
					}
				} else if (token.equals("\"")) {
					readString(fr);
				} else if (token.equals("[")) {
					ignoreSection(fr);
				} else if (token.equals("]")) {
					end = true;
					int map = mappings.size();
					if (filter) {
						if (passing.contains(new Integer(tmp_nodeid))) {
							mappings.add(tmp_nodelabel);
							itemtable.put(new Integer(tmp_nodeid), new Integer(map));
						}
						else {
							System.out.println(tmp_nodeid+ " does not pass filter");
						}
					}
					else {
						mappings.add(tmp_nodelabel);
						itemtable.put(new Integer(tmp_nodeid), new Integer(map));
					}
					//log.debug("[node id=" + tmp_nodeid + " map=" + map
					//		+ ", label=" + tmp_nodelabel + "]");
				}
			}
		}
	}

	/**
	 * @param fr
	 */
	private static void ignoreSection(BufferedReader fr) {
		boolean end = false;
		String token = "";
		while (!end) {
			token = getNextToken(fr);
			if (token.equals("\"")) {
				readString(fr);
			} else if (token.equals(("["))) {
				ignoreSection(fr);
			} else if (token.equals("]")) {
				end = true;
			}
		}

	}

	private static String readString(BufferedReader fr) {
		StringBuffer b = new StringBuffer();
		int c;
		try {
			c = fr.read();
			while (((char) c) != '"') {
				b.append((char) c);
				c = fr.read();
			}
		} catch (IOException e) {

		}
		return b.toString();

	}

	private static String getNextToken(BufferedReader fr) {
		// other string in there
		if (lastChar > -1) {
			String s = "" + (char) lastChar;
			lastChar = -1;
			return s;
		}

		StringBuffer token = new StringBuffer();
		int c = -1;

		boolean bgn = false;

		while (!bgn) {
			try {
				c = fr.read();
				if (c == -1) {
					return "";
				} else if ((c == ' ') || (c == '\t') || (c == '\n')
						|| (c == '\r')) {

				} else if ((c == '\"') || (c == '\'') || (c == '[')
						|| (c == ']')) {
					return "" + (char) c;
				} else {
					token.append((char) c);
					bgn = true;
				}
			} catch (IOException e) {
			}

		}

		boolean end = false;

		while (!end) {
			try {
				c = fr.read();
				if ((c == ' ') || (c == '\t') || (c == '\n') || (c == '\r')
						|| (c == -1)) {
					end = true;
				} else if ((c == '\"') || (c == '\'') || (c == '[')
						|| (c == ']')) {
					lastChar = c;
					end = true;
				} else {
					token.append((char) c);
				}
			} catch (IOException e) {

			}
		}

		return token.toString();
	}

}