/*
 * $Id: RevisionItem.java,v 1.5 2004/12/28 21:47:25 zimmerth Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.ccvs.model;

import java.io.Serializable;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Timestamp;
import java.util.Date;

import org.eclipse.core.runtime.Status;
import org.softevo.rose.ccvs.CCVSPlugin;

/**
* This class represents a single revision of a file in the log output.
 * <P>
 * It consists of a 
 * <UL>
 * <LI> <EM>revision</EM>, e.g., <TT>1.5</TT>.
 * <LI> <EM>author</EM>, e.g., <TT>egamma</TT>.
 * <LI> <EM>message</EM>, e.g., <TT>This is a foo log message.</TT>
 * <LI> <EM>timestamp</EM>, the time the revision was created.
 * <LI> <EM>plusLine</EM>, the number of added lines.
 * <LI> <EM>minusLines</EM>, the number of removed lines.
 * <LI> <EM>state</EM>, usually <TT>Exp</TT>, 
 * 		if the file was deleted, the <EM>state</EM> is <TT>dead</TT>.
 * <LI> <EM>position</EM>, the position of this revision in the log output.
 * </UL>
 * The affected file is directly stored in the LogInfo object.
 * 
 * @author Thomas Zimmermann, zimmerth@cs.uni-sb.de
 */
public class RevisionItem implements Serializable {

    /** The serialization ID */
	private static final long serialVersionUID = 9007329459739568493L;

	/** The identifier for this revision, e.g., <TT>1.5</TT>. */
	private String revision = null;
	/** The person who commited this revision. */
	private String author = null;
	/** The message that describes the rational of the developer. */
	private String message = null;
	
	/** The timestamp when this revision was created (in the CVS repository). */
	private Date date = null;
	/** The number of added lines. */
	private int plusLines = -1;
	/** The number of removed lines. */
	private int minusLines = -1;
	/** The state of this revision. */ 
	private String state = null;

	/** The position of this revision in the log output. */
	private int position = -1;
	
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String linesep = System.getProperty("line.separator");

		return "Revision " + revision + "  [ "
			+ "pos = #" + position + ", "
			+ "+" + plusLines+ "/" + minusLines + " lines, "
			+ "state = " + state + " ]" + linesep
			//+ "  date = " + timestamp.toString()
			+ "  author = " + author + linesep + message;
	}

    /**
     * @return Returns the author.
     */
    public String getAuthor() {
        return author;
    }

    /**
     * @return Returns the branchPrefix
     */
    public String getBranchPrefix() {
		// Determine branch
		String branch = revision.replaceFirst("\\.[0-9]+\\z", "");
		if (branch.indexOf(".") < 0) {
			branch = null;
		}
		return branch;
    }
    
    /**
     * @return Returns the message.
     */
    public String getMessage() {
        return message;
    }
    
    /**
     * @return Returns the messageMD5.
     */
    public String getMessageMD5() {
        try {
	        MessageDigest md5 = MessageDigest.getInstance("MD5");
	        md5.update(message.getBytes());
	        return RevisionItem.byte2hex(md5.digest());
	    } catch (NoSuchAlgorithmException e) {
	        CCVSPlugin.logException(e, Status.WARNING);
	    }
	    return null;
    }


    /**
     * @return Returns the minusLines.
     */
    public int getMinusLines() {
        return minusLines;
    }

    /**
     * @return Returns the plusLines.
     */
    public int getPlusLines() {
        return plusLines;
    }

    /**
     * @return Returns the position.
     */
    public int getPosition() {
        return position;
    }

    /**
     * @return Returns the revision.
     */
    public String getRevision() {
        return revision;
    }

    /**
     * @return Returns the state.
     */
    public String getState() {
        return state;
    }

    /**
     * @return Returns the timestamp.
     */
    public Date getDate() {
        return date;
    }

    /**
     * @return Returns the timestamp.
     */
    public Timestamp getTimestamp() {
        return new Timestamp(date.getTime());
    }

    /**
     * @param author The author to set.
     */
    public void setAuthor(String author) {
        this.author = author;
    }

    /**
     * @param message The message to set.
     */
    public void setMessage(String message) {
        this.message = message;
    }

    /**
     * @param minusLines The minusLines to set.
     */
    public void setMinusLines(int minusLines) {
        this.minusLines = minusLines;
    }

    /**
     * @param plusLines The plusLines to set.
     */
    public void setPlusLines(int plusLines) {
        this.plusLines = plusLines;
    }

    /**
     * @param position The position to set.
     */
    public void setPosition(int position) {
        this.position = position;
    }

    /**
     * @param revision The revision to set.
     */
    public void setRevision(String revision) {
        this.revision = revision;
    }

    /**
     * @param state The state to set.
     */
    public void setState(String state) {
        this.state = state;
    }

    /**
     * @param timestamp The timestamp to set.
     */
    public void setDate(Date timestamp) {
        this.date = timestamp;
    }
    
    /**
     * 
     * @param b
     * @return
     */
    public static String byte2hex(byte[] b) {
        String hs = "";
        String stmp = "";
        for (int n = 0; n < b.length; n++) {
            stmp = (java.lang.Integer.toHexString(b[n] & 0XFF));
            if (stmp.length() == 1)
                hs = hs + "0" + stmp;
            else
                hs = hs + stmp;
        }
        return hs.toUpperCase();
    }
}
