/*
 * $Id: EntitiesJob.java,v 1.14 2005/03/01 12:55:26 zimmerth Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.data.entities;

import java.io.IOException;
import java.io.InputStream;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.compare.IStreamContentAccessor;
import org.eclipse.compare.ResourceNode;
import org.eclipse.compare.structuremergeviewer.IStructureComparator;
import org.eclipse.compare.structuremergeviewer.IStructureCreator;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.softevo.rose.ccvs.IRepositoryProvider;
import org.softevo.rose.ccvs.RepositoryException;
import org.softevo.rose.compare.ModificationItem;
import org.softevo.rose.compare.RoseComparePlugin;
import org.softevo.rose.data.entities.db.DatabaseOperations;
import org.softevo.rose.data.entities.db.Tables;
import org.softevo.rose.data.entities.model.ChangedEntity;
import org.softevo.rose.data.entities.model.Entity;


/**
 * @author Thomas Zimmermann, zimmerth@cs.uni-sb.de
 */
public class EntitiesJob extends Job {

    public static final int WORK = 700;
    
    private DatabaseOperations dbOperations;
    private IRepositoryProvider cvs;

    /**
     * @param name
     */
    public EntitiesJob(DatabaseOperations dbOperations, IRepositoryProvider cvs) {
        super("Parse revisions for fine-grained changes");
        this.dbOperations = dbOperations;
        this.cvs = cvs;
    }

    
    /**
     * @see org.eclipse.core.runtime.jobs.Job#run(org.eclipse.core.runtime.IProgressMonitor)
     */
    protected IStatus run(final IProgressMonitor monitor) {
        try {
            
            monitor.beginTask("Detect fine-grained changes", WORK);

            // Insert all files as entities
            insertFilesAsEntities();
            
            // Get for each file a graph that is a collection of the branches
            TreeMap fileRevisionGraphs = getFileRevisionGraphs();

            // Get maximal entityID 
            int maxEntityID = -1;	
            String query = "SELECT Max(EntityID) FROM " + dbOperations.getTableName(Tables.TABLE_ENTITIES);
            
            ResultSet rset = dbOperations.executeQuery(query);
            rset.next();
            maxEntityID = rset.getInt(1);
            
            // Get the workspace root
            IContainer cvsParent = null;

            SubProgressMonitor pm = new SubProgressMonitor(monitor, WORK);
            pm.beginTask("Analyzing files", fileRevisionGraphs.keySet().size());
            
            // Sorry for this label: files
            file:
                for (Iterator fileIter = fileRevisionGraphs.keySet().iterator(); fileIter.hasNext() && !pm.isCanceled();) {
                    
                    // Get filename, fileid, and revisionList
                    SimpleFile file = (SimpleFile) fileIter.next();
                    String fileName = file.getFileName();
                    int fileID = file.getFileID();
                    List branchList = (List) fileRevisionGraphs.get(file);
                    IFile cvsFile = null;
                    
                    pm.subTask(fileName);
                    pm.worked(1);
                    
                    // Get structureCreator
                    IStructureCreator creator = RoseComparePlugin.getStructureCreator(file.getExtension());
                    if (creator!=null && creator.getName().equals("Jar Archive Compare")) creator = null;
                    if (creator!=null && creator.getName().equals("Zip Archive Compare")) creator = null;
                    boolean parseable = (creator != null);
                    
                    IStructureComparator emptyStructure = parseable ? creator.getStructure(new EmptyContent()) : null;
                    HashMap entities = new HashMap();
                    
                    // Iterate over all branches
                    branch:
                        for (Iterator branchIter = branchList.iterator(); branchIter.hasNext() && !pm.isCanceled();) {
                            SimpleBranch branch = (SimpleBranch) branchIter.next();
                            String revisionL = null;
                            String stateL = null;
                            IFile fileL = null;
                            IStructureComparator structureL = null;
                            
                            // Iterate through revisionList to determine the deltas
                            revision:
                                for (int i = 0 ; i < branch.size() && !pm.isCanceled(); i++) {
                                    
                                    // Checkout file
                                    SimpleRevision simpleRevision = (SimpleRevision) branch.get(i);
                                    
                                    String revisionR = simpleRevision.getRevision();
                                    String stateR = simpleRevision.getState();
                                    IFile fileR = null;
                                    IStructureComparator structureR = null;
                                    
                                    pm.subTask(fileName + "::" + revisionR);
                                    EntitiesPlugin.debugInfo(fileName + "::" + revisionR);
                                    
                                    // Parse the revision
                                    if (parseable && !"dead".equals(stateR)) { // parseable
                                        try {
                                            // Checkout file
                                            pm.subTask(fileName + "::" + revisionR + " checkout");
                                            String cvsFileName = fileName.startsWith("/") ? fileName.substring(1) : fileName;
                                            cvsFile = cvs.checkout(cvsFileName, revisionR, null);
                                            // Unshare old parent
                                            pm.subTask(fileName + "::" + revisionR + " unshare");
                                            if (cvsParent == null) { cvsParent = cvsFile.getParent(); }
                                            if (cvsParent != null && !cvsParent.equals(cvsFile.getParent())) {
                                                IContainer oldParent = cvsParent;
                                                cvsParent = cvsFile.getParent();
                                                IPath cvsParentPath = cvsParent.getFullPath();
                                                try {
                                                    while (!oldParent.getFullPath().isPrefixOf(cvsParentPath)) {
                                                        cvs.unshareResource(oldParent, null);
                                                        oldParent.delete(true /*force*/, null /*progress*/);
                                                        oldParent = oldParent.getParent();
                                                    }
                                                } catch (RepositoryException e) {
                                                    EntitiesPlugin.logException(e, Status.WARNING);
                                                }
                                            }
                                            // Copy file to file__revision
                                            pm.subTask(fileName + "::" + revisionR + " move");
                                            Path targetPath = new Path(cvsFile.getFullPath() + "__" + revisionR);
                                            try {
                                                cvs.unshareResource(cvsFile, null);
                                                cvsFile.move(targetPath, true /*force*/, null /*no progress*/);
                                                cvsFile.create(new EmptyInputStream(), true /*force*/, null);
                                            } catch (RepositoryException e) {
                                                EntitiesPlugin.logException(e, Status.WARNING);
                                                cvsFile.copy(targetPath, true /*force*/, null /*no progress*/);
                                            }
                                            // Parse revision
                                            pm.subTask(fileName + "::" + revisionR + " parse");
                                            fileR = ResourcesPlugin.getWorkspace().getRoot().getFile(targetPath);
                                            structureR = creator.getStructure(new ResourceNode(fileR));
                                        } catch (RepositoryException e) {
                                            EntitiesPlugin.logException(e, Status.WARNING);
                                        } catch (CoreException e) {
                                            EntitiesPlugin.logException(e, Status.WARNING);
                                        }
                                    } else { //empty
                                    }
                                    
                                    //System.out.println("free=" + Runtime.getRuntime().freeMemory());
                                    
                                    pm.subTask(fileName + "::" + revisionR + " diff/insert");
                                    LinkedList changes = new LinkedList();
                                    if (i==0) { // Special tretement for the first revision
                                        if (revisionR.equals("1.1") && stateR.equals("Exp")) {
                                            // The revision added a file
                                            // System.out.println("New file " + fileName + "::" + revisionR);
                                            if (emptyStructure != null && structureR != null) {
                                                changes = RoseComparePlugin.doStructureCompare(emptyStructure, structureR);
                                            }
                                            ChangedEntity newFile = new ChangedEntity(fileID, revisionR, fileID, ChangedEntity.ADDED);
                                            dbOperations.insertChangedEntity(newFile);
                                            
                                        } else if (revisionR.equals("1.1") && stateR.equals("dead")) {
                                            // The revision added a file but on a branch
                                            // System.out.println("New file (but on branch)" + fileName + "::" + revisionR);
                                            // Ignore until further notice
                                            
                                        } else {
                                            // The revision is the first on a branch and *likely* no addition
                                            // System.out.println("First revision on a branch " + fileName + "::" + revisionR);
                                            // Exceptions: 1.1::dead; parallel add: branch first head second
                                            // Ignore until further notice
                                        }
                                    } else { // "foo".equals(bar) does not throw an exception in case bar is null
                                        if ("dead".equals(stateL) && "dead".equals(stateR)) {
                                            // Something strange
                                            EntitiesPlugin.debugInfo("Two subsequent dead states for " + fileName + "::" + revisionL + "::" + revisionR);
                                            
                                        } else if ("dead".equals(stateL) && "Exp".equals(stateR)) {
                                            // Revision re-added a file (or a file added on a branch is merged to another)
                                            // System.out.println("Re-added (or merged file) " + fileName + "::" + revisionR);
                                            if (emptyStructure != null && structureR != null) {
                                                changes = RoseComparePlugin.doStructureCompare(emptyStructure, structureR);
                                            }
                                            ChangedEntity rebornFile = new ChangedEntity(fileID, revisionR, fileID, ChangedEntity.ADDED);
                                            dbOperations.insertChangedEntity(rebornFile);
                                            
                                        } else if ("Exp".equals(stateL) && "dead".equals(stateR)) { 
                                            // Deleted a file.
                                            // System.out.println("Deleted a file " + fileName + "::" + revisionR);
                                            if (structureL != null && emptyStructure != null) {
                                                changes = RoseComparePlugin.doStructureCompare(structureL, emptyStructure);
                                            }
                                            ChangedEntity deletedFile = new ChangedEntity(fileID, revisionR, fileID, ChangedEntity.DELETED);
                                            deletedFile.setPreviousRevisionId(revisionL);
                                            dbOperations.insertChangedEntity(deletedFile);
                                            
                                        } else if ("Exp".equals(stateL) && "Exp".equals(stateR)) { 
                                            // Changed a file.
                                            // System.out.println("Changed a file " + fileName + "::" + revisionR);
                                            if (structureL != null && structureR != null) {
                                                changes = RoseComparePlugin.doStructureCompare(structureL, structureR);
                                            }
                                            String action = parseable ? ChangedEntity.CHANGED_PARSEABLE_FILE : ChangedEntity.CHANGED_NON_PARSEABLE_FILE;
                                            ChangedEntity changedFile = new ChangedEntity(fileID, revisionR, fileID, action);
                                            changedFile.setPreviousRevisionId(revisionL);
                                            dbOperations.insertChangedEntity(changedFile);
                                        } 
                                    }
                                    
                                    for (Iterator iter = changes.iterator(); iter.hasNext();) {
                                        ModificationItem item = (ModificationItem) iter.next();
                                        
                                        // Get entityId and create entity if it doesn't exist
                                        // NOTE: For re-born entities the old entityId is used.
                                        int entityid = -1;
                                        if (!entities.containsKey(item.getID())) {
                                            if (item.getType() != ModificationItem.ADDED) {
                                                EntitiesPlugin.logWarning("WARNING: " + item.getID() + " has been changed/deleted before it was created.", null);
                                            }
                                            entityid = ++maxEntityID;
                                            Entity entity = new Entity(entityid, item.getID(), fileID);
                                            entities.put(item.getID(), new Integer(entityid));
                                            dbOperations.insertEntity(entity);
                                        } else {
                                            entityid = ((Integer) entities.get(item.getID())).intValue();
                                        }
                                        
                                        // Create changed entity
                                        String action = ChangedEntity.getCorrespondingAction(item.getType());
                                        ChangedEntity changedEntity = new ChangedEntity(fileID, revisionR, entityid, action);
                                        changedEntity.setPreviousRevisionId(revisionL);
                                        changedEntity.setContent(item.getContent());                                
                                        dbOperations.insertChangedEntity(changedEntity);
                                        
                                        if (item.isDocumentRangeNode()) {
                                            //System.out.println(item.getTypeDescription() + " " + item.getDocumentRangeNode().getId());
                                        }
                                    }
                                    
                                    // Delete fileL
                                    if (fileL != null) {
                                        fileL.delete(true /*force*/, false /*no history*/, null /*progress*/);
                                    }
                                    // Set revision1 = revision2
                                    revisionL = revisionR;
                                    stateL = stateR;
                                    fileL = fileR;
                                    structureL = structureR;
                                    
                                    
                                } // end revision loop
                            
                            // Delete fileL
                            if (fileL != null) {
                                fileL.delete(true /*force*/, false /*no history*/, null /*progress*/);
                            }
                            
                        } // end branch loop
                    
                    // Delete (empty) cvsFile
                    if (cvsFile!=null) cvsFile.delete(true /*force*/, false /*no history*/, null /*progress*/);
                    
                } // end file loop
            
        } catch (SQLException e) {
            EntitiesPlugin.logException(e, Status.WARNING);
        } catch (CoreException e) {
            EntitiesPlugin.logException(e, Status.WARNING);
        } 
        if (monitor.isCanceled()) return Status.CANCEL_STATUS;
        return Status.OK_STATUS;
    }

        
    
    
	/**
     * @return
	 * @throws SQLException
     */
    private void insertFilesAsEntities() throws SQLException {
		// Initialize entities table with file entities
		String query = "INSERT INTO " + dbOperations.getTableName(Tables.TABLE_ENTITIES) + " "
			+ "SELECT FileID, QualifiedFileName, FileID " 
			+ "FROM " + Tables.TABLE_FILES;
		dbOperations.executeUpdate(query);
    }


    /**
     * @return
	 * @throws SQLException
     */
    private TreeMap getFileRevisionGraphs() throws SQLException {
		TreeMap result = new TreeMap();
		String query = 
			"SELECT f.QualifiedFileName AS Name, f.FileID As FileID, RevisionID, BranchPrefix AS Branch, State "
			+ "FROM " + Tables.TABLE_REVISIONS + " AS r, " + Tables.TABLE_FILES + " AS f "
			+ "WHERE f.FileID=r.FileId "
			+ "ORDER BY f.QualifiedFileName, BranchPrefix DESC, LogPosition DESC;";

		ResultSet rset = dbOperations.executeQuery(query);

		String fileName = "/"; // File names don't contain a single '/'
		String branch = "";

		SimpleBranch simpleBranch = null;
		LinkedList simpleBranchList = null;
		
		// Group result by file name
		while (rset.next()) {
			// Branch can be null, so de-null it
			String rsBranch = rset.getString("Branch");
			rsBranch = rsBranch == null ? "" : rsBranch;

			if (!fileName.equals(rset.getString("Name"))) {
				// Found a new file group - create new entry in filesMap
				fileName = rset.getString("Name");
				simpleBranchList = new LinkedList();
				result.put(new SimpleFile(fileName, rset.getInt("FileID")), simpleBranchList);
				// Create a new entry in revisionLists
				branch = rsBranch;
				simpleBranch = new SimpleBranch();
				simpleBranchList.addLast(simpleBranch);

			} else if (!branch.equals(rsBranch)) {
				// Found a new branch group - create a new revisions list
				branch = rsBranch;
				simpleBranch = new SimpleBranch();
				simpleBranchList.addLast(simpleBranch);
			}
			
			// In any case, insert revision
			simpleBranch.addRevision(new SimpleRevision
			        (rset.getString("RevisionID"), rset.getString("State")));

		}
		rset.close();
		return result;
    }




    protected class SimpleFile implements Comparable {
	    
        private Pattern p = Pattern.compile("(.*/)((.+)(\\.[^.]*)|(\\.?[^.]+))");
	    private String fileName;
        private int fileID;

        public SimpleFile(String fileName, int fileID) {
	        this.fileName = fileName;
	        this.fileID = fileID;
	    }
        
        public String getFileName() {
            return fileName;
        }
        public int getFileID() {
            return fileID;
        }
        public String getExtension() {
            Matcher m = p.matcher(fileName);    		
	    		if (m.matches()) {
	    			return m.group(4);
	    		}
	    		return null;
        }

        /* (non-Javadoc)
         * @see java.lang.Comparable#compareTo(java.lang.Object)
         */
        public int compareTo(Object o) {
            SimpleFile sf = (SimpleFile) o;
            if (fileName.equals(sf.getFileName())) {
                return new Integer(fileID).compareTo(new Integer(sf.getFileID()));
            }
            return fileName.compareTo(sf.getFileName());
        }        
	}
    
	protected class SimpleRevision {
	    
	    private String revision;
        private String state;

        public SimpleRevision(String revision, String state) {
	        this.revision = revision;
	        this.state = state;
	    }
        
        public String getRevision() {
            return revision;
        }
        public String getState() {
            return state;
        }
	}

	protected class SimpleBranch extends LinkedList {

	    private static final long serialVersionUID = 6009335074727417445L;

	    public void addRevision(SimpleRevision revision) {
	        addLast(revision);
	    }
	}

	private class EmptyContent implements IStreamContentAccessor {
	    
        /**
         * @see org.eclipse.compare.IStreamContentAccessor#getContents()
         */
        public InputStream getContents() throws CoreException {
            return new EmptyInputStream();
        }
	    
	}
	
	private class EmptyInputStream extends InputStream {
        public int read() throws IOException {
            // TODO Auto-generated method stub
            return -1;
        }	    
	}
	
}
