/*
 * $Id: ExtractJob.java,v 1.8.2.1 2005/04/09 12:43:48 zimmerth Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.data.mirror;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.TreeSet;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.softevo.rose.ccvs.IRepositoryProvider;
import org.softevo.rose.ccvs.RepositoryException;
import org.softevo.rose.ccvs.listener.AbstractLogOutputListener;
import org.softevo.rose.ccvs.listener.IRepositoryContentListener;
import org.softevo.rose.ccvs.model.BranchItem;
import org.softevo.rose.ccvs.model.LogInfo;
import org.softevo.rose.ccvs.model.RevisionItem;
import org.softevo.rose.ccvs.model.TagItem;
import org.softevo.rose.ccvs.parser.DefaultLogParser;
import org.softevo.rose.data.mirror.db.DatabaseOperations;
import org.softevo.rose.data.mirror.db.Tables;

/**
 * @author Thomas Zimmermann, zimmerth@cs.uni-sb.de
 */
public class ExtractJob extends Job {

    public static final int WORK = 200;
    
    private DatabaseOperations dbOperations;
    private IRepositoryProvider cvs;

    /**
     * 
     * @param dbOperations
     * @param cvs
     */
    public ExtractJob(DatabaseOperations dbOperations, IRepositoryProvider cvs) {
        super("Mirror CVS repository into database");
        this.dbOperations = dbOperations;
        this.cvs = cvs;
    }

    /**
     * @see org.eclipse.core.runtime.jobs.Job#run(org.eclipse.core.runtime.IProgressMonitor)
     */
    protected IStatus run(final IProgressMonitor monitor) {
        try {
            monitor.beginTask("Mirror repository", WORK);
            final SubProgressMonitor pm = new SubProgressMonitor(monitor, 25);
            pm.beginTask("Get repository content", IProgressMonitor.UNKNOWN);
            final MyInteger files = new MyInteger();
            final DatabaseOperations db = dbOperations;
            cvs.getRepositoryContent(new IRepositoryContentListener() {
                public void doFile(String fileName) {
                    files.i++;
                    monitor.subTask(files.i + " files");
                    monitor.worked(1);
                }
                public void doDeletedFile(String fileName) {
                    files.i++;
                    pm.subTask(files.i + " files");
                    pm.worked(1);
                }
                public void doDirectory(String directoryName) {
                    // Moved to Log Output (Bug #1, reported by Lance Norskog)
                    //try {
                    //    db.insertDirectory(directoryName);
                    //} catch (SQLException e) {
                    //    MirrorPlugin.logException(e, Status.WARNING);
                    //}
                }                        
            }, pm);
            pm.done();
            
            final MyInteger fileID = new MyInteger(getFileId());
            final MyInteger transactionID = new MyInteger(getTransaktionId());
            
            final SubProgressMonitor pm2 = new SubProgressMonitor(monitor, WORK - 25);
            pm2.beginTask("Get log messages", files.i);
            cvs.getLogs(new AbstractLogOutputListener() {
                	private TreeSet directories = new TreeSet();
	                public void doBranch(BranchItem branch) {
	                    try {
                            db.insertBranch(fileID.i, branch);
                        } catch (SQLException e) {
                            pm2.setCanceled(true);
                            MirrorPlugin.logException(e, Status.WARNING);
                        }
	                }
	
	                public void doLog(LogInfo logInfo) {
	                    try {
	                        pm2.subTask(logInfo.getFileName());
	                        pm2.worked(1);
	                        // Determine (and insert) directory 
	                        String directoryName = logInfo.getDirectoryName();
	                        if (!directories.contains(directoryName)) {
	                            try {
	                                db.insertDirectory(directoryName);
	                            } catch (SQLException e) {
	                                pm2.setCanceled(true);
	                                MirrorPlugin.logException(e, Status.WARNING);
	                            }
	                            directories.add(directoryName);
	                        }
	                        // Insert file
	                        db.insertFile(++fileID.i, logInfo);
                        } catch (SQLException e) {
                            pm2.setCanceled(true);
                            MirrorPlugin.logException(e, Status.WARNING);
                        }
	                }
	
	                public void doRevision(RevisionItem revision) {
	                    try {
                            db.insertTransaktion(++transactionID.i, revision);
                            db.insertRevision(transactionID.i, fileID.i, revision);
                        } catch (SQLException e) {
                            pm2.setCanceled(true);
                            MirrorPlugin.logException(e, Status.WARNING);
                        }
	                }
	
	                public void doTag(TagItem tag) {
	                    try {
                            db.insertTag(fileID.i, tag);
                        } catch (SQLException e) {
                            pm2.setCanceled(true);
                            MirrorPlugin.logException(e, Status.WARNING);
                        }
	                }
	            }, 
            	new DefaultLogParser(), pm2);
            pm2.done();
            cvs.close();
        } catch (RepositoryException e) {
            MirrorPlugin.logException(e, Status.WARNING);
        }
        if (monitor.isCanceled()) return Status.CANCEL_STATUS;
        return Status.OK_STATUS;
    }

	/**
     * @return the maximal transaction id
     */
    private int getTransaktionId() {
        String transactionTable = dbOperations.getTableName(Tables.TABLE_TRANSACTIONS);
        int maxTransaktionID = 0;
		try {
            String query = "SELECT MAX(TransaktionID) FROM " + transactionTable;
            ResultSet rset = dbOperations.executeQuery(query);
            rset.next();
            maxTransaktionID = rset.getInt(1);
            rset.close();
        } catch (SQLException e) {
            // ignore
        }
        return maxTransaktionID;
    }

    /**
     * @return the maximal file id
     */
    private int getFileId() {
        String fileTable = dbOperations.getTableName(Tables.TABLE_FILES);
        int maxFileID = 0;
		try {
            String query = "SELECT MAX(FileID) FROM " + fileTable;
            ResultSet rset = dbOperations.executeQuery(query);
            rset.next();
            maxFileID = rset.getInt(1);
            rset.close();
        } catch (SQLException e) {
            // ignore
        }
        return maxFileID;
    }

    private class MyInteger {
	    public int i = 0;
	    public MyInteger() {}
	    public MyInteger(int i) { this.i = i; }	    
	}

}
