/*
 * $Id: DBConnectionPropertiesFactory.java,v 1.2 2005/01/12 16:06:13 dallmeier Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.db;

import java.util.Properties;

import org.softevo.rose.db.jdbc.DatabaseProperties;
import org.softevo.rose.db.jdbc.JdbcPlugin;

/** This class is responsible for the creation
 *  of properties objects with type information. <BR>
 *  Its intention is to separate serialization information
 *  from connection state. Changes in the type hierarchy of
 *  the <code>DBConnectionProperties</code> class should be
 *  reflected here.
 *  @author Valentin Dallmeier
 */
public class DBConnectionPropertiesFactory {

  /** Key for type attribute. <BR>
   */
  public static final String KEY_TYPE = "CONNECTION_TYPE";
  
  /** String constant for default connections type. <BR>
   */
  public static final String TYPE_DEFAULT = "default";
  
  /** String constant for internal connections type. <BR>
   */
  public static final String TYPE_INTERNAL = "internal";

  /** This method tries to create a connection properties
   *  object from the properties passed. <BR>
   *  A type identifier is expected to be found in the properties.
   *  If none is present, this method assumes <code>DBConnectionProperties</code>
   *  as default type. 
   *  @param properties the properties
   *  @return the connection properties
   */
  public static DBConnectionProperties loadFromProperties(Properties properties) {
    String type;
    DBConnectionProperties connectionProperties;
    
    type = properties.getProperty(KEY_TYPE);
    if (type != null) {
      if (type.equals(TYPE_DEFAULT)) {
        connectionProperties = new DBConnectionProperties(properties);
        connectionProperties.setDatabaseProperties(JdbcPlugin.getDefault().getProperties(properties.getProperty(DatabaseProperties.KEY_NAME)));
        return connectionProperties;
      } else if (type.equals(TYPE_INTERNAL)) {
        return new InternalDBConnectionProperties();
      }
    }
    return new DBConnectionProperties(properties);
  }
  
  /** Creates properties for the connection properties
   *  passed. <BR>
   *  This method writes a type identifier in addition to
   *  the values already saved by the connection properties. 
   *  @param connectionProperties the connection properties
   *  @return the properties that were created
   */
  public static Properties createProperties(DBConnectionProperties connectionProperties) {
    Properties result;
    
    result = new Properties();
    connectionProperties.setProperties(result);
    if (connectionProperties instanceof InternalDBConnectionProperties) {
      result.put(KEY_TYPE, TYPE_INTERNAL);
    } else if (connectionProperties instanceof DBConnectionProperties) {
      result.put(KEY_TYPE, TYPE_DEFAULT);
    } else {
      //if the type is unknown, use a default type
      result.put(KEY_TYPE, TYPE_DEFAULT);
    }
    return result;
  }
}
