/*
 * $Id: DatabaseProvider.java,v 1.9 2005/01/30 09:38:04 dallmeier Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.HashSet;


/** This class provides access to a database. <BR>
 *  A <code>DatabaseProvider</code> maintains a sql connection
 *  and must be opened and closed. <BR>
 */
public class DatabaseProvider {

  /** The set of users. <BR>
   */
  private HashSet users = null;
  
  /** A synchronization lock. <BR>
   */
  private Object lock = null;
  
  /** The connection to the database. <BR>
   */
  private Connection connection = null;

  /** The database connection properties. <BR>
   */
  private DBConnectionProperties connectionProperties = null;

  
  /** Creates a new database provider. <BR>
   *  @param connectionProperties the properties describing the connection
   */
  public DatabaseProvider(DBConnectionProperties connectionProperties) {
    this.connectionProperties = connectionProperties;
    users = new HashSet();
    lock = new Object();
  }
  
  /** Opens a connection to the database. <BR>
   *  If the provider was opened before, this method
   *  does nothing.
   *  @throws SQLException if an exception occurs
   *          connecting to the database
   */
  public void connect() throws SQLException {
    if (connection == null) {
      try {
        Class.forName(connectionProperties.getDatabaseProperties().getJdbcDriverClassName());
      } catch (Exception instantiationException) {
        throw new SQLException("Unable to load jdbc driver " + 
                               connectionProperties.getDatabaseProperties().getJdbcDriverClassName() + ":" +
                               instantiationException.getClass().getName() +
                               ":" + instantiationException.getMessage());
      }
      connection = DriverManager.getConnection(connectionProperties.getDatabaseURL(),
                                               connectionProperties.getUserName(),
                                               connectionProperties.getPassword());
    }
  }
  
  /** Closes the connection to the database. <BR>
   *  Every subsequent call to <code>getConnection</code> 
   *  fails. If the provider was not opened before,
   *  this method does nothing.
   *  @throws SQLException if an exception occurs disconnecting
   */
  public void close() throws SQLException {
    if (connection != null) {
      try {
        connection.close();
      } finally {
        connection = null;
      }
    }
  }
  
  /** Gets the connection. <BR>
   *  If the provider was not initialized before,
   *  this method simply returns <code>null</code>.
   *  @return the connection of this provider
   */
  public Connection getConnection() {
    return connection;
  }
  
  /** Gets the connection properties. <BR>
   */
  public DBConnectionProperties getConnectionProperties() {
    return connectionProperties;
  }

  /** Tests wether or not the connection was
   *  opened before. <BR>
   *  Note that this method only tests if the
   *  connection was created before. No checks
   *  are performed on sql level.
   *  @return <code>true</code> if the connection is open,
   *          <code>false</code> otherwise
   */
  public boolean isConnected() {
    return connection != null;
  }
  
  /** Deregisters a previously registered user object
   *  from the provider. <BR>
   *  This method raises a <code>RuntimeException</code>
   *  if the object was not registered before.
   *  @param user the user object
   */
  public void deRegisterUser(Object user) {
    if (users.contains(user)) {
      users.remove(user);
      if (users.size() == 0) {
        synchronized(lock) {
          lock.notify(); // FIXME This doesn't work: Thread not owner!!!
        }
      }
    } else {
      throw new RuntimeException("FATAL: Attempt to remove user object from provider " +
                                 "not registered before");
    }
  }
  
  /** Registers an object as user in the provider. <BR>
   *  @param user the user object
   */
  public void registerUser(Object user) {
    users.add(user);
  }
  
  /** This method waits for a free provider. <BR>
   *  The implementation guarantees that after
   *  execution of this method, no users are registered
   *  in the provider. Note that this method is not protected
   *  by a timeout.
   */
  public void waitForFreeProvider() {
    while (users.size() > 0) {
      try {
        synchronized (lock) {
          lock.wait();
        }
      } catch (InterruptedException wakeUp) {
        //ignore exception
      }
    }
  }
}
