/*
 * $Id: JdbcPlugin.java,v 1.12 2005/03/01 12:54:26 zimmerth Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.db.jdbc;

import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.Vector;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;
import org.osgi.framework.BundleContext;

/**
 * This plugin provides access to a set of configured
 * jdbc drivers. <BR>
 */
public class JdbcPlugin extends Plugin {
  
  /** String constant for drivers extension point. <BR>
   */
  public static final String EXT_DRIVERS = "org.softevo.rose.db.jdbc.drivers";
  
  /** String constant for driver tag. <BR>
   */
  public static final String TAG_DRIVER = "driver";
  
  /** String constant for mapping tag. <BR>
   */
  public static final String TAG_MAPPING = "mapping";
  
  /** String constant for value tag. <BR>
   */
  public static final String TAG_ID = "id";

  /** String constant for driver class attribute. <BR>
   */
  public static final String ATT_DRIVERCLASS = "class";
  
  /** String constant for type class attribute. <BR>
   */
  public static final String ATT_TYPECLASS = "class";
  
  /** String constant for type value attribute. <BR>
   */
  public static final String ATT_TYPEVALUE = "value";
  
  /** String constant for id value attribute. <BR>
   */
  public static final String ATT_IDVALUE = "value";

  /** The name of the driver for the internal database. <BR>
   */
  public static final String NAME_INTERNALDRIVER = "hsqldb";

  /** The list of database properties read from the database. <BR>
   */
  private Vector databaseProperties = null;
  
  
  /** Singleton instance. <BR>
   */
  private static JdbcPlugin plugin;
  
  /** Resource bundle for plugin. <BR>
   */
  private ResourceBundle resourceBundle;
  
  /**
   * The constructor.
   */
  public JdbcPlugin() {
    super();
    plugin = this;
    try {
      resourceBundle = ResourceBundle.getBundle("org.softevo.rose.db.jdbc.JdbcPluginResources");
    } catch (MissingResourceException x) {
      resourceBundle = null;
    }
  }
  
  /**
   * This method is called upon plug-in activation
   */
  public void start(BundleContext context) throws Exception {
    super.start(context);
  }
  
  /**
   * This method is called when the plug-in is stopped
   */
  public void stop(BundleContext context) throws Exception {
    super.stop(context);
  }
  
  /**
   * Returns the shared instance.
   */
  public static JdbcPlugin getDefault() {
    return plugin;
  }
  
  /**
   * Returns the string from the plugin's resource bundle,
   * or 'key' if not found.
   */
  public static String getResourceString(String key) {
    ResourceBundle bundle = JdbcPlugin.getDefault().getResourceBundle();
    try {
      return (bundle != null) ? bundle.getString(key) : key;
    } catch (MissingResourceException e) {
      return key;
    }
  }
  
  /**
   * Returns the plugin's resource bundle,
   */
  public ResourceBundle getResourceBundle() {
    return resourceBundle;
  }
  
  /** Gets the database properties stored in
   *  the plugin. <BR>
   *  The objects in the list must not be modified.
   */
  public Vector getDatabaseProperties() {
    Vector copy;
    DatabaseProperties properties;
    
    if (databaseProperties == null) {
      databaseProperties = readDrivers();
    }
    copy = new Vector(databaseProperties.size());
    for (int counter = 0; counter < databaseProperties.size(); counter++) {
      properties = (DatabaseProperties) databaseProperties.get(counter);
      copy.add(properties.clone());
    }
    return copy;
  }
  
  /** Tests if the driver is loadable or not. <BR>
   *  @param properties the properties describing the driver
   *  @return <code>true</code> if the driver is loadable, 
   *          <code>false</code> otherwise
   */
  protected boolean testDriver(DatabaseProperties properties) {
    try {
      Class.forName(properties.getJdbcDriverClassName());
      return true;
    } catch (Exception loadException) {
      getLog().log(new Status(IStatus.WARNING, "org.softevo.rose.db", 0, loadException.getMessage(), loadException));
      return false;
    }
  }

  /** Gets the database properties for a given name. <BR>
   *  @param name the name of the properties sought
   *  @return the properties or <code>null</code> if none were 
   *          found for the name passed
   */
  public DatabaseProperties getProperties(String name) {
    DatabaseProperties result;
    Vector properties;
    
    properties = getDatabaseProperties();
    for (int counter = 0; counter < properties.size(); counter++) {
      result = (DatabaseProperties) properties.get(counter);
      if (result.getName().equals(name)) {
        return (DatabaseProperties) result.clone();
      }
    }
    return null;
  }
  /** Reads the driver configuration lazily from
   *  the plugin manifest. <BR>
   *  @return a set of <code>DatabaseProperties</code> for all drivers configured
   */
  protected Vector readDrivers() {
    IExtensionRegistry registry;
    IExtensionPoint extensionPoint;
    IExtension[] extensions;
    IConfigurationElement[] elements;
    Vector result;
    DatabaseProperties properties;
    
    registry = Platform.getExtensionRegistry();
    extensionPoint = registry.getExtensionPoint(EXT_DRIVERS);
    extensions = extensionPoint.getExtensions();
    result = new Vector(extensions.length);
    for (int counter = 0; counter < extensions.length; counter++) {
      properties = new DatabaseProperties(null, null);
      elements = extensions[counter].getConfigurationElements();
      for (int counter2 = 0; counter2 < elements.length; counter2++) {
        if (elements[counter2].getName().toLowerCase().equals(TAG_DRIVER)) {
          properties.setJdbcDriverClassName(elements[counter2].getAttribute(ATT_DRIVERCLASS));
        } else if (elements[counter2].getName().toLowerCase().equals(TAG_MAPPING)) {
          properties.setTypeMapping(elements[counter2].getAttribute(ATT_TYPECLASS),
                                    elements[counter2].getAttribute(ATT_TYPEVALUE));
        } else if (elements[counter2].getName().toLowerCase().equals(TAG_ID)) {
          properties.setName(elements[counter2].getAttribute(ATT_IDVALUE));
        }
      }
      if (testDriver(properties)) {
        result.add(properties);
      }
    }
    return result;
  }
  
}
