/*
 * $Id: DBConnectionsView.java,v 1.10.2.1 2005/04/09 12:43:48 zimmerth Exp $
 * 
 * LICENSE:
 */
package org.softevo.rose.db.preferences;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.window.Window;
import org.eclipse.jface.wizard.IWizard;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.ui.internal.Workbench;
import org.eclipse.ui.model.WorkbenchContentProvider;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.softevo.rose.db.DBConnectionProperties;
import org.softevo.rose.db.DbPlugin;
import org.softevo.rose.db.wizards.EditDBConnectionWizard;

/** A tree view that presents all connections
 *  currently configured and buttons that
 *  bring up an editor for the currently selected
 *  connection or to create a new instance. <BR>
 */
public class DBConnectionsView {
  
  /** String constant for connection editing. <BR>
   */
  private static final String TITLE_EDIT_LOCATION = "Edit connection";
  
  /** The parent of this view. <BR>
   */
  private Composite parent = null;
  
  /** The viewer for the connections tree. <BR>
   */
  private TreeViewer treeViewer = null;
  
  /** The root group containing all the elements. <BR>
   */
  private Group rootGroup = null;
  
  /** Button to create new connection. <BR>
   */
  private Button newButton = null;
  
  /** Button to delete connection. <BR>
   */
  private Button deleteButton = null;
  
  /** Button to edit connection. <BR>
   */
  private Button editButton = null;
  
  /** The currently selected element. <BR>
   */
  private DBConnectionProperties selection = null;

  private IProject project = null;

  private IWizard wizard = null;
  
  /** Creates the connections view. <BR>
   *  @param parent the editor parent
   */
  public DBConnectionsView(Composite parent) {
    this.parent = parent;
    createElements();
  }
  
  public DBConnectionsView(Composite parent, IProject project) {
    this.parent = parent;
    this.project = project;
    createElements();
  }

  /** Gets the currently selected properties element. <BR>
   *  A return value of <code>null</code> indicates that
   *  no element is currently selected. 
   */
  public DBConnectionProperties getSelection() {
    return selection;
  }
  
  /** Creates the view elements. <BR>
   */
  private void createElements() {
    rootGroup = createRootGroup();
    treeViewer = createDBConnectionsTree(rootGroup);
    addButtons(rootGroup);
  }
  
  /** Creates the group that contains all other elements. <BR>
   */
  private Group createRootGroup() {
    GridData gridData;
    Group result;
    GridLayout layout;
    
    gridData = new GridData(GridData.GRAB_HORIZONTAL | 
                            GridData.HORIZONTAL_ALIGN_FILL | 
                            GridData.VERTICAL_ALIGN_FILL);
    gridData.horizontalSpan = 2;
    gridData.heightHint = 300;
    result = new Group(parent, SWT.SHADOW_ETCHED_IN);
    result.setLayoutData(gridData);
    result.setText("Database Locations:");
    layout = new GridLayout(1, false);
    result.setLayout(layout);
    return result;
  }

  /** This method tries to extract the connection properties
   *  from the selection passed. <BR>
   *  @param structuredSelection the selection that contains the <code>ConnectionElement</code>
   *  @return the properties object or <code>null</code>
   */
  private DBConnectionProperties getPropertiesFromSelection(IStructuredSelection structuredSelection) {
    if (structuredSelection.isEmpty() || !(structuredSelection.getFirstElement() instanceof ConnectionElement)) {
      return null;
    }
    return ((ConnectionElement)structuredSelection.getFirstElement()).getProperties();
  }

  /** This method may be called to select a certain properties
   *  object. <BR>
   *  @param properties the properties to select
   */
  public void selectPropertiesInTree(DBConnectionProperties properties) {
    StructuredSelection structuredSelection;
    
    structuredSelection = new StructuredSelection(new ConnectionElement(properties));
    updateButtons(properties);
    treeViewer.setSelection(structuredSelection, true);
    selection = properties;
  }
  
  /** Updates the tree view. <BR>
   */
  public void updateTreeView() {
    treeViewer.refresh();
  }
  
  /** Creates the tree viewer. <BR>
   */
  private TreeViewer createDBConnectionsTree(Composite parent) {
    Tree tree;
    GridData gridData;
    TreeViewer result;
    
    tree = new Tree(parent, SWT.SINGLE | SWT.BORDER);
    tree.setBackground(DbPlugin.getDefault().getColorRegistry().get(DbPlugin.COLOR_WHITE));
    gridData = new GridData(GridData.FILL_BOTH);
    tree.setLayoutData(gridData);
    gridData.heightHint = 150;
    result = new TreeViewer(tree);
    result.setContentProvider(new WorkbenchContentProvider());
    result.setLabelProvider(new WorkbenchLabelProvider());
    result.addSelectionChangedListener(new ISelectionChangedListener() {
      public void selectionChanged(SelectionChangedEvent event) {
        IStructuredSelection structuredSelection;
        
        structuredSelection = (IStructuredSelection)event.getSelection();
        selection = getPropertiesFromSelection(structuredSelection);
        updateButtons(selection);
      }
    });
    result.setInput(new ConnectionsElement(project));
    return result;
  }

  /** Changes the enabled state of the edit
   *  and delete buttons as needed by the
   *  properties object passed (typically the selection). <BR>
   *  @param properties the properties
   */
  private void updateButtons(DBConnectionProperties properties) {
    if (properties != null) {
      editButton.setEnabled(properties.isEditable());
      deleteButton.setEnabled(properties.isEditable());
    } else {
      editButton.setEnabled(false);
      deleteButton.setEnabled(false);
    }
    if (wizard!=null) wizard.getContainer().updateButtons();
  }

  /** Brings up the connection wizard. <BR>
   *  @param properties the properties object. <BR>
   *  @param title the title for the dialog
   *  @return the properties that were edited or created
   */
  private DBConnectionProperties showConnectionWizard(DBConnectionProperties properties, String title) {
    EditDBConnectionWizard wizard;
    WizardDialog dialog;
    
    wizard = new EditDBConnectionWizard(properties, title);
    dialog = new WizardDialog(Workbench.getInstance().getActiveWorkbenchWindow().getShell(), wizard);
    if (dialog.open() == Window.OK) {
      return wizard.getEditedProperties();
    } 
    return null;
  }
  
  /** Adds the edit, new and delete buttons. <BR>
   *  @param parent the visual parent
   */
  private void addButtons(Composite parent) {
    Composite composite;
    FillLayout fillLayout;
    ErrorDialog errorDialog;
    
    composite = new Composite(parent, SWT.NULL);
    fillLayout = new FillLayout();
    composite.setLayout(fillLayout);
    newButton = new Button(composite, SWT.PUSH);
    newButton.setText("New...");
    newButton.addSelectionListener(new SelectionListener() {
      public void widgetSelected(SelectionEvent event) {
        ISelection selectionData;
        selectionData = treeViewer.getSelection();
        DBConnectionProperties result = showConnectionWizard(null, "Create new location");
        if (result != null) {
          selection = result;
        }
        treeViewer.refresh();
        treeViewer.setSelection(selectionData, true);
        updateButtons(getPropertiesFromSelection((IStructuredSelection) selectionData));
      }
      public void widgetDefaultSelected(SelectionEvent event) {
        widgetSelected(event);
      }
    });
    editButton = new Button(composite, SWT.PUSH);
    editButton.setText("Edit...");
    editButton.addSelectionListener(new SelectionListener() {
      public void widgetSelected(SelectionEvent event) {
        ISelection selectionData;
        if (selection != null) {
          selectionData = treeViewer.getSelection();
          DBConnectionProperties result = showConnectionWizard(selection, TITLE_EDIT_LOCATION + " " + selection.getName());
          if (result != null) {
            selection = result;
          }
          treeViewer.refresh();
          editButton.setEnabled(selection.isEditable());
        }
      }
      public void widgetDefaultSelected(SelectionEvent event) {
        widgetSelected(event);
      }
    });
    //if nothing is selected, disable edit button
    if (selection == null) {
      editButton.setEnabled(false);
    }
    deleteButton = new Button(composite, SWT.PUSH);
    deleteButton.setText("Delete");
    deleteButton.setEnabled(false);
    deleteButton.addSelectionListener(new SelectionListener() {
      public void widgetSelected(SelectionEvent event) {
        if (selection != null) {
          ErrorDialog errorDialog;
          
          try {
            DbPlugin.getDefault().removeDatabaseConnectionProperties(selection.getName());
          } catch (Exception exception) {
            errorDialog = new ErrorDialog(Workbench.getInstance().getActiveWorkbenchWindow().getShell(),
                                          exception.getClass().getName(), 
                                          "An exception occured deleting connection " + 
                                          selection.getName() + ": " + exception.getMessage(),
                                          Status.CANCEL_STATUS, Status.CANCEL);
            errorDialog.open();
          }
          treeViewer.refresh();
          editButton.setEnabled(false);
          selection = null;
        }
      }
      public void widgetDefaultSelected(SelectionEvent event) {
        widgetSelected(event);
      }
    });
  }

/**
 * @param wizard
 */
public void setWizard(IWizard wizard) {
    this.wizard = wizard;
}
  
}
