/*
 * $Id: FileUtils.java,v 1.3 2005/03/01 10:42:30 zimmerth Exp $
 * 
 * LICENSE:
 */
package org.softevo.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;

/**
 * Fileutil provides some useful operations for handling of files.
 * 
 * @author zimmerth
 */
public class FileUtils {

	/**
	 * Copies a file. If both files are identical nothing is coped, but
	 * <code>true</code> is returned. Directories containing the target file 
	 * are created if they don't exist yet. If an error occurs during copy
	 * <code>false</code> is returned.
	 * 
	 * @param src  The source file.
	 * @param dest  The target file.
	 * @return  <code>true</code> if the copy operation was successfull.
	 */
	public static boolean copy(File src, File dest) {

		boolean result = true;
		
		FileInputStream fis = null;
		FileOutputStream fos = null;

		try {
			// Don't copy, if files are equal
			if (src.getCanonicalFile().equals(dest.getCanonicalFile())) {
				return true; 
			}
			
			// Create superdirectories of target
			if (dest.getParentFile() != null) {
				dest.getParentFile().mkdirs();
			}

			// Copy
			fis = new FileInputStream(src);
			fos = new FileOutputStream(dest);

			byte[] buffer = new byte[8 * 1024];
			int i = 0;
			while ((i = fis.read(buffer)) != -1) {
				fos.write(buffer, 0, i);
			}

		} catch (IOException e) {
			// Copy operation failed
			result = false;

		} finally {
			// Close streams.			
			try {
				if (fis != null) { fis.close(); } 
			} catch (Exception e) { e.printStackTrace(); }
			try {
				if (fos != null) { fos.close();	} 
			} catch (Exception e) { e.printStackTrace(); }
		}
		
		return result;
	}
	
    /**
     * Writes a String to a file overriding the file if it does exist.
     *
     * @param fileName 
     * @param data 
     * @param encoding 
     * @throws IOException 
     */
    public static void writeStringToFile(String fileName, String data, String encoding) throws IOException {
        File file = new File(fileName);
        if (file.exists()) {
            file.delete();
        }
        OutputStream out = new FileOutputStream(file);
        try {
            out.write(data.getBytes(encoding));
        } finally {
            try {
                out.close();
            } catch (IOException e) {
                // ignore
            }
        }
    }

    
    /**
     * <p>
     * Reads the contents of a file into a String.
     * </p>
     * <p>
     * There is no readFileToString method without encoding parameter because
     * the default encoding can differ between platforms and therefore results
     * in inconsistent results.
     * </p>
     *
     * @param fileName  
     * @param encoding  
     * @return String
     * @throws IOException 
     */
    public static String readFileToString(String fileName, String encoding) throws IOException {
        StringBuffer data = new StringBuffer(4096);
        InputStreamReader input = new InputStreamReader(new FileInputStream(fileName), encoding);
        try {
            char[] buffer = new char[4096];
            int n = 0;
            while (-1 != (n = input.read(buffer))) {
                data.append(buffer, 0, n);
            }
        } finally {
            try {
                input.close();
            } catch (IOException e) {
                // ignore
            }
        }
        return data.toString();
    }
    
}
